/*
 * Copyright (c) 2008, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef _API_LIB_H_
#define _API_LIB_H_

/*
 * Definitions used by the OpenFC-specific library for the SNIA HBA-API.
 */
#define	HBA_API_VENDOR        "Open-FC.org"
#define	HBA_API_VERSION       "2.2"
#define	HBA_API_VENDOR_RURL   "org.open-fc"	/* reversed URL */

#endif /* _API_LIB_H_ */
